﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text.RegularExpressions;
using System.Xml.Linq;

namespace UniversalMigrator.Core
{
	public static class DirectoryExtensions
	{
		public static void CopyFolder(string sourceFolder, string destinationFolder)
		{
			Directory.CreateDirectory(destinationFolder);

			foreach (string file in Directory.GetFiles(sourceFolder, "*.*", SearchOption.AllDirectories))
			{
				string fileName = Path.GetFileName(file);
				string filePath = Path.GetDirectoryName(file.Substring(sourceFolder.Length + 1));

				string destFilePath;

				if (!string.IsNullOrEmpty(filePath))
				{
					string childFolder = Path.Combine(destinationFolder, filePath);

					Directory.CreateDirectory(childFolder);

					destFilePath = Path.Combine(childFolder, fileName);
				}
				else
				{
					destFilePath = Path.Combine(destinationFolder, fileName);
				}

				File.Copy(file, destFilePath, true);
			}
		}
	}

	public interface SkinChange
	{
		void Apply(ref string css, string skin);
	}

	public class StringRewrite : SkinChange
	{
		public string From { get; set; }
		public string To { get; set; }

		public void Apply(ref string css, string skin)
		{
			css.Replace(
				String.Format(From, skin),
				String.Format(To, skin));
		}
	}

	public class AppendSelectors : SkinChange
	{
		private static Regex skinNameReferences = new Regex("{skin}", RegexOptions.Compiled | RegexOptions.Multiline | RegexOptions.IgnoreCase);

		public string Selectors { get; set; }

		public void Apply(ref string css, string skin)
		{
			css += skinNameReferences.Replace(Selectors, ((Match m) => skin));
		}
	}

	public static class MigrationEngine
	{
		private static Regex releasePattern = new Regex("(?<year>20\\d\\d)(\\.|\\s)q(?<quarter>[1-3])", RegexOptions.Compiled | RegexOptions.IgnoreCase);
		private static Regex skinCssFilenamePattern = new Regex("(?<control>[a-z]+)\\.(?<skin>[a-z0-9_\\-]+).css", RegexOptions.Compiled | RegexOptions.IgnoreCase);

		public static void Migrate(string sourceFolder, string destinationFolder, string fromRelease, string toRelease)
		{
			if (!Directory.Exists(sourceFolder))
				throw new ArgumentOutOfRangeException("sourceFolder", "Source folder does not exist");

			if (!releasePattern.IsMatch(fromRelease))
				throw new ArgumentOutOfRangeException("fromRelease", "Parameter should be in format {year}.q{1-3}, starting from 2009.q2");

			if (!releasePattern.IsMatch(toRelease))
				throw new ArgumentOutOfRangeException("toRelease", "Parameter should be in format {year}.q{1-3}, starting from 2009.q2");

			if (destinationFolder != sourceFolder)
			{
				if (Directory.Exists(destinationFolder))
				{
					foreach (var file in Directory.GetFiles(destinationFolder, "*.*", SearchOption.AllDirectories))
					{
						File.SetAttributes(file, FileAttributes.Normal);
						File.Delete(file);
					}

					Directory.Delete(destinationFolder, true);
				}

				DirectoryExtensions.CopyFolder(sourceFolder, destinationFolder);
			}

			var accumulatedChanges = new Dictionary<string, List<SkinChange>>();

			do
			{
				fromRelease = GetReleaseString(fromRelease, GetNextRelease);

				AccumulateSkinChanges(fromRelease, ref accumulatedChanges);

			} while (fromRelease != GetReleaseString(toRelease, null));

			ApplySkinChanges(destinationFolder, accumulatedChanges);
		}

		public delegate void ReleaseModificationCallback (ref int year, ref int quarter);

		public static string GetReleaseString(string release, ReleaseModificationCallback callback)
		{
			Match releaseMatch = releasePattern.Match(release);

			var year = Convert.ToInt32(releaseMatch.Groups["year"].Value);
			var quarter = Convert.ToInt32(releaseMatch.Groups["quarter"].Value);

			if (callback != null)
			{
				callback(ref year, ref quarter);
			}

			return String.Format("{0}.q{1}", year, quarter);
		}

		public static void GetNextRelease(ref int year, ref int quarter)
		{
			quarter++;

			if (quarter == 4) {
				quarter = 1;
				year++;
			}
		}

		public static void AccumulateSkinChanges(string release, ref Dictionary<string, List<SkinChange>> accumulatedChanges)
		{
			var resourcesAssembly = Assembly.GetExecutingAssembly();

			var migrationDocument = String.Format("UniversalMigrator.Core.MigrationDocuments.{0}.xml", release.ToUpper());

			using (var migrationDocumentStream = resourcesAssembly.GetManifestResourceStream(migrationDocument))
			{
				if (migrationDocumentStream == null)
					return;

				using (var imageStreamReader = new StreamReader(migrationDocumentStream))
				{
					var doc = XDocument.Load(imageStreamReader).Elements();

					foreach (var skinChangeTypes in doc.Descendants())
					{
						foreach (var control in skinChangeTypes.Descendants())
						{
							if (!accumulatedChanges.Keys.Contains(control.Name.LocalName))
							{
								accumulatedChanges.Add(control.Name.LocalName, new List<SkinChange>());
							}

							accumulatedChanges[control.Name.LocalName].Add(new AppendSelectors() { Selectors = control.Value });
						}
					}
				}
			}
		}

		private static void ApplySkinChanges(string destinationFolder, Dictionary<string, List<SkinChange>> accumulatedChanges)
		{
			foreach (string file in Directory.GetFiles(destinationFolder, "*.css", SearchOption.AllDirectories))
			{
				string fileName = Path.GetFileName(file);

				if (!skinCssFilenamePattern.IsMatch(fileName))
					continue;

				Match skinFilename = skinCssFilenamePattern.Match(fileName);

				var control = skinFilename.Groups["control"].Value;
				var skin = skinFilename.Groups["skin"].Value;

				var cssText = File.ReadAllText(file);

				if (accumulatedChanges.ContainsKey(control))
				{
					foreach (var change in accumulatedChanges[control])
					{
						change.Apply(ref cssText, skin);
					}
				}

				File.SetAttributes(file, FileAttributes.Normal);

				File.WriteAllText(file, cssText);
			}
		}
	}
}